/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bp2fr
 *
 * bp2fr is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bp2fr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <bpipe/bpipe.h>
#include <bpipe/datatypes.h>
#include <bpipe/bpipe_axaf.h>

#include <fullray/fullray.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"

void
process(Global *g)
{
  FILE *fout;
  BPipe *bpipe;
  fullray fr;

  size_t s_obj_in, nphot;

  void *obj_in;

  DVector3 *	position;
  DVector3 *	direction;
  DVector3 *	surface_norm;
  double *	weight;
  double *	graze_angle;
  double *	energy;
  double *	path_length;
  DCVector2 *	pol_cos;
  DCVector2 *	pol_sin;
  unsigned int *	id;
  int *		handy;
  DpktField *field;   
  BPMatrix *bpmatrix; 
  
  tf_enter("process");

  if (! strcmp(g->output, "stdout") )
    fout = stdout;
  else
    {
      fout = fopen(g->output, "w");
      if (NULL == fout)
	tf_exit(ExitERR_fopen, "couldn't open %s for output", g->output);
    }
  
  if ( NULL == ( bpipe = bpipe_new( ) ) )
    tf_exit(ExitERR_alloc, "unable to create bpipe" );
  
  if ( bpipe_input( bpipe, g->input ) )
    tf_exit( ExitERR_alloc, "unable to attach input stream `%s' to bpipe: %s",
             g->input, bpipe_strerror( bpipe_errno ) );
  
  /* create missing (non graze angle) fields in data packet */
  if (
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_phot_pos,	BPDType_DVector3, NULL )
    ||
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_phot_dir,	BPDType_DVector3, NULL )
    ||
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_surf_norm,	BPDType_DVector3, NULL )
    ||
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_phot_wt,	BPDType_double,   NULL )
    ||
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_phot_energy,	BPDType_double,	NULL )
    ||
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_path_length,	BPDType_double, NULL )
    ||
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_phot_id,	BPDType_uint,	NULL )
    ||
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_pol_cos,	BPDType_DCVector2,   NULL)
    ||
    -1 == bpipe_dpktf_add( bpipe,
			   BP_AXAF_pol_sin,	BPDType_DCVector2,   NULL)
    ||
    -1 == bpipe_dpktf_array_add(bpipe,
				"handy",		BPDType_int,	3)
    )
    tf_die( "error creating data missing packet field: %s",
	    bpipe_strerror( bpipe_errno) );
  
  /* graze angle field in data packet must have extent of 2 */
  
  if (NULL == ( field = bpipe_dpktf( bpipe, BP_AXAF_phot_graze_angle)))
    /* field does not exist, so create with extent 2 */
    bpipe_dpktf_array_add( bpipe,
			   BP_AXAF_phot_graze_angle, BPDType_double, 2 );
  else
    {
      /* need to resize with extent 2 */      
      if ( NULL == (bpmatrix = bpipe_dpktf_matrix(field, BPDSite_CORE, NULL )))
	{
	  tf_die("bpipe_dpktf_matrix failure: %s", 
		 bpipe_strerror( bpipe_errno ));
	}
      bpmatrix->extent[0] = 2;
      if (bpipe_dpktf_resize_core( field, bpmatrix, NULL, NULL, NULL))
	{
	  tf_die("bpipe_dpktf_resize_core failure: %s", 
		 bpipe_strerror( bpipe_errno ));
	}
    }
  
  if ( 0 == ( s_obj_in = bpipe_map(bpipe) ) )
    tf_die( "no data packet fields on input stream" );
  
  if ( NULL == (obj_in = malloc(s_obj_in)) )
    tf_exit(ExitERR_alloc, "unable to allocate object(s)");
  
  position	= (DVector3 *)	dpktf_data( bpipe, BP_AXAF_phot_pos, obj_in );
  direction	= (DVector3 *)	dpktf_data( bpipe, BP_AXAF_phot_dir, obj_in );
  surface_norm	= (DVector3 *)	dpktf_data( bpipe, BP_AXAF_surf_norm, obj_in );
  weight	= (double *)	dpktf_data( bpipe, BP_AXAF_phot_wt, obj_in );
  energy	= (double *)	dpktf_data( bpipe, BP_AXAF_phot_energy,
					    obj_in );
  path_length	= (double *)	dpktf_data( bpipe, BP_AXAF_path_length,
					    obj_in );
  graze_angle	= (double *)	dpktf_data( bpipe, BP_AXAF_phot_graze_angle,
					    obj_in );
  id		= (unsigned int *) dpktf_data( bpipe, BP_AXAF_phot_id,
					       obj_in );
  handy		= (int *)	dpktf_data(bpipe, "handy", obj_in );
  pol_cos	= (DCVector2 *)	dpktf_data( bpipe, BP_AXAF_pol_cos, obj_in );
  pol_sin	= (DCVector2 *)	dpktf_data( bpipe, BP_AXAF_pol_sin, obj_in );

  /*
    preset data in case not all fields are available
    from the bpipe stream
   */
  position->x = 0.0;
  position->y = 0.0;
  position->z = 0.0;

  direction->x = 0.0;
  direction->y = 0.0;
  direction->z = 0.0;

  surface_norm->x = 0.0;
  surface_norm->y = 0.0;
  surface_norm->z = 0.0;

  *weight = 0.0;

  graze_angle[0] = 0.0;
  graze_angle[1] = 0.0;

  *energy = 0.0;
  *path_length = 0.0;

  pol_cos->q1.r = 0.0;
  pol_cos->q1.i = 0.0;
  pol_cos->q2.r = 0.0;
  pol_cos->q2.i = 0.0;

  pol_sin->q1.r = 0.0;
  pol_sin->q1.i = 0.0;
  pol_sin->q2.r = 0.0;
  pol_sin->q2.i = 0.0;

  *id = 0;

  handy[0] = 0;
  handy[1] = 0;
  handy[2] = 0;

  for ( nphot = 1 ; bpipe_read_dpkts( bpipe, obj_in, 1) == 1 ; nphot++ )
  {
    fr.pos.r[0] = position->x;
    fr.pos.r[1] = position->y;
    fr.pos.r[2] = position->z;

    fr.pos.v[0] = direction->x;
    fr.pos.v[1] = direction->y;
    fr.pos.v[2] = direction->z;

    fr.norm[0] = surface_norm->x;
    fr.norm[1] = surface_norm->y;
    fr.norm[2] = surface_norm->z;

    fr.wt = weight[0];

    fr.g_ang[0] = graze_angle[1];
    fr.g_ang[1] = graze_angle[0];

    fr.energy = energy[0];
    fr.path = path_length[0];

    fr.c2[0].r = pol_cos->q1.r;
    fr.c2[0].i = pol_cos->q1.i;
    fr.c2[1].r = pol_cos->q2.r;
    fr.c2[1].i = pol_cos->q2.i;
		    
    fr.s2[0].r = pol_sin->q1.r;
    fr.s2[0].i = pol_sin->q1.i;
    fr.s2[1].r = pol_sin->q2.r;
    fr.s2[1].i = pol_sin->q2.i;

    fr.id = id[0];

    fr.handy[0] = handy[0];
    fr.handy[1] = handy[1];
    fr.handy[2] = handy[2];

    if ( fwrite(&fr, sizeof(fullray), 1, fout) != 1)
      tf_exit(ExitERR_fwrite, "error writing photon %d to output", nphot );
  }

  fclose( fout );

  if ( bpipe_errno == BPEIOERR )
      tf_exit(ExitERR_fwrite, "error reading photon %d from input channel",
	      nphot );

  bpipe_delete( bpipe );

  free( obj_in );

  tf_leave( );
}

